/*
 * This file is part of musicd.
 * Copyright (C) 2011-2012 Konsta Kokkinen <kray@tsundere.fi>
 * 
 * Musicd is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * Musicd is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with Musicd.  If not, see <http://www.gnu.org/licenses/>.
 */
#include "protocol_http.h"

#include "cache.h"
#include "client.h"
#include "config.h"
#include "image.h"
#include "json.h"
#include "log.h"
#include "musicd.h"
#include "query.h"
#include "strings.h"
#include "task.h"

#include "library.h"

typedef struct http {
  client_t *client;

  struct {
    stream_t *stream;
    AVFormatContext *avctx;
    AVIOContext *avioctx;
    uint8_t *aviobuf;
  } stream;
} http_t;

static const char *get_ptr(const char *args, const char *key)
{
  const char *p = args;

  if (!args) {
    return NULL;
  }

  while (1) {
    if (strbeginswith(p, key)) {
      return p + strlen(key);
    }
    
    p = strchr(p, '&');
    if (!p) {
      break;
    }
    ++p;
  }
  return NULL;
}

static int64_t get_int(const char *args, const char *key)
{
  const char *p = get_ptr(args, key);
  int64_t result = 0;

  if (!p) {
    return 0;
  }

  /* The parameter is set but it has no value */
  if (*p != '=') {
    return 0;
  }
  
  sscanf(p, "=%" PRId64 "", &result);
  
  return result;
}

static bool get_bool(const char *args, const char *key)
{
  const char *p = get_ptr(args, key);
  return p ? true : false;
}

static char *get_str(const char *args, const char *key)
{
  const char *p = get_ptr(args, key);
  string_t *result;
  char tmp;

  if (!p) {
    return NULL;
  }
  
  /* The parameter is set but it has no value */
  if (*p != '=') {
    return strdup("");
  }
  
  result = string_new();
  for (++p; *p != '&' && *p != '\0'; ++p) {
    if (*p != '%') {
      if (*p == '+') { /* + means space */
        string_push_back(result, ' ');
      } else {
        string_push_back(result, *p);
      }
      continue;
    }
    
    ++p;

    if (!isxdigit(*p) || !isxdigit(*(p+1)) || sscanf(p, "%2hhx", &tmp) != 1) {
      string_free(result);
      return NULL;
    }
    string_push_back(result, tmp);
    
    ++p;
  }
  
  return string_release(result);
}

/**
 * @param status default 200 OK if NULL
 * @param content_type default text/html if NULL
 */
static void http_send
  (http_t *http,
   const char *status,
   const char *content_type,
   size_t content_length,
   const char *content)
{
  client_send(http->client,
              "HTTP/1.1 %s\r\n"
              "Server: musicd/" MUSICD_VERSION_STRING "\r\n"
              "Content-Type: %s; charset=utf-8\r\n"
              "Content-Length: %d\r\n"
              "Access-Control-Allow-Origin: *\r\n"
              "\r\n",
              status ? status : "200 OK",
              content_type ? content_type : "text/html",
              content_length);
  client_write(http->client, content, content_length);
}

static void http_send_text
  (http_t *http,
   const char *status,
   const char *content_type,
   const char *content)
{
  http_send(http, status, content_type, strlen(content), content);
}

static void http_reply(http_t *http, const char *status)
{
  client_send(http->client,
              "HTTP/1.1 %s\r\n"
              "Server: musicd/" MUSICD_VERSION_STRING "\r\n"
              "Content-Type: text/plain; charset=utf-8\r\n"
              "Content-Length: %d\r\n"
              "Access-Control-Allow-Origin: *\r\n"
              "\r\n%s", status, strlen(status), status);
}

static void http_send_file
  (http_t *http, const char *path, const char *content_type)
{
  size_t size;
  char *data;
  FILE *file = fopen(path, "rb");
  if (!file) {
    http_reply(http, "404 Not Found");
  }

  fseek(file, 0, SEEK_END);
  size = ftell(file);
  fseek(file, 0, SEEK_SET);

  if (size <= 0) {
    http_reply(http, "404 Not Found");
    fclose(file);
    return;
  }

  data = malloc(size);
  size = fread(data, 1, size, file);
  fclose(file);

  http_send(http, NULL, content_type, size, data);

  free(data);
}

static char *decode_url_value(const char **p)
{
  char tmp;
  string_t *result = string_new();

  for (; **p != '&' && **p != '\0'; ++*p) {
    if (**p != '%') {
      if (**p == '+') { /* + means space */
        string_push_back(result, ' ');
      } else {
        string_push_back(result, **p);
      }
      continue;
    }    
    ++*p;

    if (!isxdigit(**p) || !isxdigit(*(*p+1)) || sscanf(*p, "%2hhx", &tmp) != 1) {
      string_free(result);
      return NULL;
    }
    string_push_back(result, tmp);
    
    ++*p;
  }
  return string_release(result);
}


/** Iterates through all arguments and sets all valid filters to query. */
static int args_to_query_filters(const char *args, query_t *query)
{
  const char *p;
  char *name, *value;
  query_field_t field;

  if (!args) {
    return 0;
  }
  p = args;
  while (1) {
    for (args = p; *p != '\0' && *p != '=' && *p != '&'; ++p) { }

    if (*p == '\0') {
      break;
    }
    if (*p == '&') {
      args = ++p;
      continue;
    }

    name = strextract(args, p);
    field = query_field_from_string(name);
    free(name);
    
    ++p;

    if (!field) {
      for (; *p != '\0' && *p != '&'; ++p) { }
    } else {
      value = decode_url_value(&p);
      if (value) {
        query_filter(query, field, value);
        free(value);
      }
    }

    if (*p == '\0') {
      break;
    }
    ++p;
  }
  return 0;
}

static void args_to_query_bounds(const char *args, query_t *query)
{
  int64_t limit = get_int(args, "limit"),
          offset = get_int(args, "offset");
  if (limit > 0) {
    query_limit(query, limit);
  }
  if (offset > 0) {
    query_offset(query, offset);
  }
}

static void args_to_query_sort(const char *args, query_t *query)
{
  char *sort = get_str(args, "sort");
  if (!sort) {
    return;
  }
  query_sort_from_string(query, sort);
  free(sort);
}

static int64_t args_to_total(const char *args, query_t *query)
{
  if (!get_bool(args, "total")) {
    return 0;
  }
  return query_count(query);
}

static int method_musicd(http_t *http, const char *args)
{
  (void)args;
  json_t json;

  json_init(&json);
  json_object_begin(&json);
  json_define(&json, "name"); json_string(&json, config_get("server-name"));
  json_define(&json, "version");  json_string(&json, MUSICD_VERSION_STRING);
  json_define(&json, "http-api"); json_string(&json, "1");

  json_define(&json, "codecs");
  json_array_begin(&json);
    json_string(&json, "mp3");
  json_array_end(&json);
  json_define(&json, "bitrate-min"); json_int(&json, 64000);
  json_define(&json, "bitrate-max"); json_int(&json, 320000);

  json_define(&json, "image-sizes");
  json_array_begin(&json);
    json_int(&json, 16);
    json_int(&json, 32);
    json_int(&json, 64);
    json_int(&json, 128);
    json_int(&json, 256);
    json_int(&json, 512);
  json_array_end(&json);

  json_object_end(&json);

  http_send_text(http, "200 OK", "text/json", json_result(&json));

  json_finish(&json);
  return 0;
}

static int method_tracks(http_t *http, const char *args)
{
  query_t *query = query_tracks_new();
  int64_t total;
  json_t json;
  track_t track;

  args_to_query_filters(args, query);

  total = args_to_total(args, query);
  if (total < 0) {
    musicd_log(LOG_ERROR, "protocol_http", "query_count failed");
    http_reply(http, "500 Internal Server Error");
    goto finish;
  }
  
  args_to_query_bounds(args, query);
  args_to_query_sort(args, query);
  
  if(query_start(query)) {
    musicd_log(LOG_ERROR, "protocol_http", "query_start failed");
    http_reply(http, "500 Internal Server Error");
    goto finish;
  }
  
  json_init(&json);
  json_object_begin(&json);
  
  if (total) {
    json_define(&json, "total");
    json_int64(&json, total);
  }
  
  json_define(&json, "tracks");
  json_array_begin(&json);
  
  while (!query_tracks_next(query, &track)) {
    json_object_begin(&json);
    json_define(&json, "id");       json_int64(&json, track.id);
    json_define(&json, "track");    json_int(&json, track.track);
    json_define(&json, "title");    json_string(&json, track.title);
    json_define(&json, "artistid"); json_int64(&json, track.artistid);
    json_define(&json, "artist");   json_string(&json, track.artist);
    json_define(&json, "albumid");  json_int64(&json, track.albumid);
    json_define(&json, "album");    json_string(&json, track.album);
    json_define(&json, "duration"); json_int(&json, track.duration);
    json_object_end(&json);
  }
  json_array_end(&json);
  json_object_end(&json);
  
  http_send_text(http, "200 OK", "text/json", json_result(&json));

  json_finish(&json);

finish:
  query_close(query);
  return 0;
}

static int method_artists(http_t *http, const char *args)
{
  query_t *query = query_artists_new();
  int64_t total;
  json_t json;
  query_artist_t artist;

  args_to_query_filters(args, query);

  total = args_to_total(args, query);
  if (total < 0) {
    musicd_log(LOG_ERROR, "protocol_http", "query_count failed");
    http_reply(http, "500 Internal Server Error");
    goto finish;
  }
  
  args_to_query_bounds(args, query);
  args_to_query_sort(args, query);
  
  if(query_start(query)) {
    musicd_log(LOG_ERROR, "protocol_http", "query_start failed");
    http_reply(http, "500 Internal Server Error");
    goto finish;
  }
  
  json_init(&json);
  json_object_begin(&json);
  
  if (total) {
    json_define(&json, "total");
    json_int64(&json, total);
  }
  
  json_define(&json, "artists");
  json_array_begin(&json);
  
  while (!query_artists_next(query, &artist)) {
    json_object_begin(&json);
    json_define(&json, "id");       json_int64(&json, artist.artistid);
    json_define(&json, "artist");    json_string(&json, artist.artist);
    json_object_end(&json);
  }
  json_array_end(&json);
  json_object_end(&json);
  
  http_send_text(http, "200 OK", "text/json", json_result(&json));
  
  json_finish(&json);

finish:
  query_close(query);
  return 0;
}

static int method_albums(http_t *http, const char *args)
{
  query_t *query = query_albums_new();
  int64_t total;
  json_t json;
  query_album_t album;

  args_to_query_filters(args, query);

  total = args_to_total(args, query);
  if (total < 0) {
    musicd_log(LOG_ERROR, "protocol_http", "query_count failed");
    http_reply(http, "500 Internal Server Error");
    goto finish;
  }
  
  args_to_query_bounds(args, query);
  args_to_query_sort(args, query);
  
  if(query_start(query)) {
    musicd_log(LOG_ERROR, "protocol_http", "query_start failed");
    http_reply(http, "500 Internal Server Error");
    goto finish;
  }
  
  json_init(&json);
  json_object_begin(&json);
  
  if (total) {
    json_define(&json, "total");
    json_int64(&json, total);
  }
  
  json_define(&json, "albums");
  json_array_begin(&json);
  
  while (!query_albums_next(query, &album)) {
    json_object_begin(&json);
    json_define(&json, "id");       json_int64(&json, album.albumid);
    json_define(&json, "album");    json_string(&json, album.album);
    json_object_end(&json);
  }
  json_array_end(&json);
  json_object_end(&json);
  
  http_send_text(http, "200 OK", "text/json", json_result(&json));

  json_finish(&json);

finish:
  query_close(query);
  return 0;
}

static int64_t validate_image_size(int64_t size)
{
  if (size == 0) {
    return 0;
  }
  if (size < 16) {
    return 16;
  }
  if (size > 512) {
    return 512;
  }
  return size;
}

static int send_image(http_t *http, char *cache_name)
{
  char *data;
  int data_size;
  data = cache_get(cache_name, &data_size);
  if (!data) {
    http_reply(http, "404 Not Found");
  } else {
    http_send(http, "200 OK", "image/jpeg", data_size, data);
  }

  free(data);
  free(cache_name);
  return 0;
}

static int handle_image(http_t *http, int64_t image, int64_t size)
{
  char *cache_name, *path;
  task_t *task;
  image_task_t *task_args;

  size = validate_image_size(size);
  if (!size) {
    path = library_image_path(image);
    if (!path) {
      http_reply(http, "404 Not Found");
      return 0;
    }
    http_send_file(http, path, image_mime_type(path));
    free(path);
    return 0;
  }

  cache_name = image_cache_name(image, size);
  if (cache_exists(cache_name)) {
    return send_image(http, cache_name);
  }

  task_args = malloc(sizeof(image_task_t));
  task_args->id = image;
  task_args->size = size;
  task = task_start(image_task, (void *)task_args);
  client_wait_task(http->client, task, (client_callback_t)send_image, cache_name);
  return 0;
}

static int method_image(http_t *http, const char *args)
{
  int64_t image, size;

  image = get_int(args, "id");
  size = get_int(args, "size");
  if (image <= 0) {
    http_reply(http, "400 Bad Request");
  }

  return handle_image(http, image, size);
}

static int method_album_image(http_t *http, const char *args)
{
  int64_t album, size, image;

  album = get_int(args, "id");
  size = get_int(args, "size");

  image = library_album_image(album);
  if (image <= 0) {
    http_reply(http, "404 Not Found");
    return 0;
  }

  return handle_image(http, image, size);
}

static bool album_images_cb(library_image_t *image, json_t *json)
{
  json_int64(json, image->id);
  return true;
}

static int method_album_images(http_t *http, const char *args)
{
  int64_t album;
  json_t json;

  album = get_int(args, "id");
  if (album <= 0) {
    http_reply(http, "400 Bad Request");
    return 0;
  }

  json_init(&json);
  json_object_begin(&json);
  json_define(&json, "images");
  json_array_begin(&json);
  library_iterate_images_by_album(album,
    (bool(*)(library_image_t *, void *))album_images_cb, &json);
  json_array_end(&json);
  json_object_end(&json);
  http_send_text(http, "200 OK", "text/json", json_result(&json));
  json_finish(&json);

  return 0;
}

static int write_packet(void *opaque, uint8_t *buf, int buf_size)
{
  http_t *http = (http_t *)opaque;
  client_write(http->client, (char *)buf, buf_size);
  return buf_size;
}

static int http_feed(void *opaque)
{
  http_t *http = (http_t *)opaque;
  uint8_t *data;
  size_t size;
  int64_t pts;
  AVPacket packet;
  int result;
  
  data = stream_next(http->stream.stream, &size, &pts);
  
  if (!data) {
    client_stop_feed(http->client);
    return -1;
  }
  
  av_init_packet(&packet);
  packet.data = data;
  packet.size = size;
  packet.pts = pts;
  packet.dts = pts;
  packet.stream_index = 0;
  
  result = av_interleaved_write_frame(http->stream.avctx, &packet);
  if (result < 0) {
    musicd_log(LOG_ERROR, "protocol_http",
               "av_interleaved_write_frame failed: %s",
               strerror(AVUNERROR(result)));
    client_stop_feed(http->client);
    return -1;
  }
  
  return 0;
}

static int method_open(http_t *http, const char *args)
{
  int result;
  int64_t id, seek, bitrate;
  track_t *track = NULL;
  stream_t *stream;
  transcoder_t *transcoder;
  AVFormatContext *avctx;
  AVStream *avstream;
  uint8_t *aviobuf;
  AVIOContext *avioctx;
  
  id = get_int(args, "id");
  seek = get_int(args, "seek");
  bitrate = get_int(args, "bitrate");
  if (!bitrate) {
    bitrate = 320000;
  } else if (bitrate < 64000) {
    bitrate = 64000;
  } else if (bitrate > 320000) {
    bitrate = 320000;
  }

  track = library_track_by_id(id);
  if (!track) {
    http_reply(http, "404 Not Found");
    return 0;
  }

  stream = stream_open(track);
  if (!stream) {
    http_reply(http, "500 Internal Server Error");
    track_free(track);
    return 0;
  }

  transcoder = transcoder_open(&stream->src_format, "mp3", bitrate);
  if (!transcoder) {
    http_reply(http, "500 Internal Server Error");
    stream_close(stream);
    return 0;
  }

  stream_set_transcoder(stream, transcoder);
  
  if (stream_seek(stream, seek) < 0) {
    http_reply(http, "500 Internal Server Error");
    stream_close(stream);
    return 0;
  }
  
  result = avformat_alloc_output_context2(&avctx, NULL, "mp3", NULL);
  if (result < 0) {
    musicd_log(LOG_ERROR, "protocol_http",
               "avformat_alloc_output_context2: %s",
               strerror(AVUNERROR(result)));
    http_reply(http, "500 Internal Server Error");
    stream_close(stream);
    return 0;
  }

  avstream =
    avformat_new_stream(avctx,
      avcodec_find_encoder_by_name(stream->format->codec));
  if (!avstream) {
    musicd_log(LOG_ERROR, "protocol_http", "avformat_new_stream failed");
    http_reply(http, "500 Internal Server Error");
    stream_close(stream);
    return 0;
  }
  
  av_dict_set(&avctx->metadata, "track", stringf("%02d", track->track),
              AV_DICT_DONT_STRDUP_VAL);
  av_dict_set(&avctx->metadata, "title", track->title, 0);
  av_dict_set(&avctx->metadata, "artist", track->artist, 0);
  av_dict_set(&avctx->metadata, "album", track->album, 0);
  
  avstream->codec = avcodec_alloc_context3(NULL);
  avcodec_copy_context(avstream->codec, stream->transcoder->encoder);
  
  aviobuf = av_mallocz(4096 * 40);
  avioctx =
    avio_alloc_context(aviobuf, 4096, 1, http, NULL, write_packet, NULL);
  if (!avioctx) {
    musicd_log(LOG_ERROR, "protocol_http", "avio_alloc_context failed");
    http_reply(http, "500 Internal Server Error");
    av_free(aviobuf);
    avformat_free_context(avctx);
    stream_close(stream);
    return 0;
  }

  avctx->pb = avioctx;

  client_send(http->client,
    "HTTP/1.1 200 OK\r\n"
    "Server: musicd/" MUSICD_VERSION_STRING "\r\n"
    "Content-Type: audio/mpeg\r\n"
    "Access-Control-Allow-Origin: *\r\n"
    "\r\n");
  
  http->stream.stream = stream;
  http->stream.avctx = avctx;
  http->stream.aviobuf = aviobuf;
  http->stream.avioctx = avioctx;
  
  avformat_write_header(http->stream.avctx, NULL);
  
  client_start_feed(http->client);

  return 0;
}


struct method_entry {
  const char *name;
  int (*handler)(http_t *http, const char *args);
};
static struct method_entry methods[] = {
  { "/musicd", method_musicd },
  { "/tracks", method_tracks },
  { "/artists", method_artists },
  { "/albums", method_albums },

  { "/image", method_image },
  { "/album/image", method_album_image },
  { "/album/images", method_album_images },

  { "/open", method_open },

  { NULL, NULL }
};

static int process_query(http_t *http, const char *path, const char *args)
{
  struct method_entry *entry;
  for (entry = methods; entry->name != NULL; ++entry) {
    if (!strcmp(entry->name, path)) {
      return entry->handler(http, args);
    }
  }

  http_reply(http, "404 Not Found");
  return 0;
}

static int http_detect(const char *buf, size_t buf_size)
{
  (void)buf_size;
  
  if (!config_to_bool("enable-http")) {
    return -1;
  }

  if (strbeginswith(buf, "GET ")
   || strbeginswith(buf, "HEAD ")) {
    return 1;
  }
  
  if (buf_size < strlen("HEAD ")) {
    return 0;
  }

  return -1;
}

static void *http_open(client_t *client)
{
  http_t *http = malloc(sizeof(http_t));
  memset(http, 0, sizeof(http_t));
  http->client = client;
  return http;
}

static void http_close(void *self)
{
  http_t *http = (http_t *)self;

  if (http->stream.stream) {
    stream_close(http->stream.stream);
  }
  if (http->stream.avctx) {
    avformat_free_context(http->stream.avctx);
  }
  if (http->stream.avioctx) {
    av_free(http->stream.avioctx);
  }
  if (http->stream.aviobuf) {
    av_free(http->stream.aviobuf);
  }
  free(http);
}

static int http_process(void *self, const char *buf, size_t buf_size)
{
  (void)buf_size;
  http_t *http = (http_t *)self;
  const char *end, *p1, *p2;
  char *query, *path, *args;
  int result = 0;

  /* Do we have all headers? */
  end = strstr(buf, "\r\n\r\n");
  if (!end) {
    /* Not enough data */
    return 0;
  }
  end += 4;

  /* Is this an HTTP method we can handle? */
  if (strbeginswith(buf, "GET ")
   || strbeginswith(buf, "HEAD ")) {
  } else {
    musicd_log(LOG_VERBOSE, "protocol_http",
               "unsupported http method (not GET or HEAD)");
    http_reply(http, "400 Bad Request");
    return -1;
  }
  
  /* Extract HTTP query */
  for (p1 = buf; *p1 != ' '; ++p1) { }
  ++p1;
  for (p2 = p1; *p2 != ' '; ++p2) {
    if (*p2 == '\0' || *p2 == '\r' || *p2 == '\n') {
      musicd_log(LOG_VERBOSE, "protocol_http",
                 "malformed request line (no tailing version)");
      musicd_log(LOG_DEBUG, "protocol_http", "request was:\n%s", buf);
      http_reply(http, "400 Bad Request");
      return -1;
    }
  }
  query = strextract(p1, p2);
  
  musicd_log(LOG_VERBOSE, "protocol_http", "query: %s", query);

  /* Extract path and arguments from query */
  p2 = strchr(query, '?');
  if (!p2) {
    /* No arguments */
    path = strdup(query);
    args = NULL;
  } else {
    path = strextract(query, p2);
    args = strextract(p2 + 1, NULL);
  }
  
  musicd_log(LOG_DEBUG, "protocol_http", "path: %s, args: %s", path, args);
  
  result = process_query(http, path, args);

  free(query);
  free(path);
  free(args);

  if (result < 0) {
    return result;
  }
  return end - buf;
}

protocol_t protocol_http = {
  .name = "http",
  .detect = http_detect,
  .open = http_open,
  .close = http_close,
  .process = http_process,
  .feed = http_feed
};

